/*	
	Copyright(c) 2010 Christian Riess <christian.riess@cs.fau.de>.

	This file may be licensed under the terms of of the GNU General Public
	License, version 3, as published by the Free Software Foundation. You can
	find it here: http://www.gnu.org/licenses/gpl.html
*/

#ifndef INDEX_H
#define INDEX_H

#include <cassert>

/** Template class for storing indices, i.e. image coordinates.
 *
 * The first template parameter is the dimension of the coordinates (two
 * dimensional images need typically two dimensional indices, x and y).
 *
 * The second template parameter is the datatype of the index. The default
 * argument unsigned short allows (on typical machines) 16 bit long index entries.
 */
template <int dimension=2, typename indexT = unsigned short>
class Index {
public:
	Index() { }
	virtual ~Index() { }
	
	Index<dimension, indexT> &operator=(Index<dimension, indexT> &idx) {
		for (int i = 0; i < dimension; ++i) {
			indices[i] = idx.get(i);
		}
		return *this;
	}

	Index<dimension, indexT> &operator=(const Index<dimension, indexT> &idx) {
		for (int i = 0; i < dimension; ++i) {
			indices[i] = idx.get(i);
		}
		return *this;
	}

	Index<dimension, indexT> *operator=(Index<dimension, indexT> *idx) {
		for (int i = 0; i < dimension; ++i) {
			(*indices)[i] = idx->get(i);
		}
		return *this;
	}

	/*
	Index<dimension, indexT> &operator=(Index<dimension, indexT> idx) {
		for (int i = 0; i < dimension; i++) {
			indices[i] = idx.get(i);
		}
		return *this;
	}
	*/
	indexT &operator[](int dim) {
		assert(dim >= 0);
		assert(dim < dimension);
		return indices[dim];
	}

	virtual indexT get(int dim) {
		assert(dim >= 0);
		assert(dim < dimension);
		return indices[dim];
	}

	virtual const indexT get(int dim) const {
		assert(dim >= 0);
		assert(dim < dimension);
		return indices[dim];
	}

	virtual void set(int dim, indexT val) {
		assert(dim >= 0);
		assert(dim < dimension);
		indices[dim] = val;
	}
protected:
	indexT indices[dimension];
};

// two dimensional index: the first dimension is the x-coordinate of the pixels
// that should be clustered, the second dimension is the index of the connected component
class Index2D : public Index<2, unsigned int> {
public:
	Index2D() {
		indices[0] = 0;
		indices[1] = 0;
	}

	Index2D(unsigned int x, unsigned int y) {
		indices[0] = x;
		indices[1] = y;
	}

	unsigned int &operator[](int dim) {
		assert(dim >= 0);
		assert(dim < 2);
		return indices[dim];
	}


	virtual ~Index2D() { }
};

namespace vole {
	typedef Index<2, unsigned short> Index2s;
	typedef Index<2, signed short> Index2ss;
	typedef Index<2, unsigned int> Index2i;
	typedef Index<2, signed int> Index2si;
}

#endif // INDEX_H
